﻿using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;
using Microsoft.Diagnostics.Runtime;
using System;

namespace MultidimensionalArrayBenchmark {
	public class Program {
		private const int Seed = 42;

		[Params(128)]
		public int Dim1 { get; set; }

		[Params(128)]
		public int Dim2 { get; set; }

		[Params(4)]
		public int Dim3 { get; set; }

		[Params(8)]
		public int Dim4 { get; set; }

		[Params(10000)]
		public int AccessCount { get; set; }

		[Params(42)] public int X { get; set; }
		[Params(13)] public int Y { get; set; }
		[Params(3)] public int D { get; set; }
		[Params(7)] public int T { get; set; }

		private int[,,,] array4;
		private int[][][][] array1111;
		private int[,][][] array211;

		[GlobalSetup]
		public void AllocArrays() {
			array4 = new int[Dim1, Dim2, Dim3, Dim4];

			array1111 = new int[Dim1][][][];
			for (int y = 0; y < array1111.Length; y++) {
				array1111[y] = new int[Dim2][][];
				for (int x = 0; x < array1111[y].Length; x++) {
					array1111[y][x] = new int[Dim3][];
					for (int d = 0; d < array1111[y][x].Length; d++) {
						array1111[y][x][d] = new int[Dim4];
					}
				}
			}

			array211 = new int[Dim1, Dim2][][];
			for (int y = 0; y < Dim1; y++) {
				for (int x = 0; x < Dim2; x++) {
					array211[y, x] = new int[Dim3][];
					for (int d = 0; d < array211[y, x].Length; d++) {
						array211[y, x][d] = new int[Dim4];
					}
				}
			}
		}

		[Benchmark]
		public int[,,,] RandomArray4() {
			var r = new Random(Seed);

			for (int i = 0; i < AccessCount; i++) {
				array4[r.Next(Dim1), r.Next(Dim2), r.Next(Dim3), r.Next(Dim4)] = 1;
			}

			return array4;
		}

		[Benchmark]
		public int[,,,] ConstantArray4() {
			for (int i = 0; i < AccessCount; i++) {
				array4[1, 1, 1, 1] = 1;
			}

			return array4;
		}

		[Benchmark]
		public int[,,,] OneArray4() {
			array4[X, Y, D, T] = 1;
			
			return array4;
		}

		[Benchmark]
		public int[][][][] RandomArray1111() {
			var r = new Random(Seed);

			for (int i = 0; i < AccessCount; i++) {
				array1111[r.Next(Dim1)][r.Next(Dim2)][r.Next(Dim3)][r.Next(Dim4)] = 1;
			}

			return array1111;
		}

		[Benchmark]
		public int[][][][] ConstantArray1111() {
			var r = new Random(Seed);

			for (int i = 0; i < AccessCount; i++) {
				array1111[1][1][1][1] = 1;
			}

			return array1111;
		}

		[Benchmark]
		public int[][][][] OneArray1111() {
			array1111[X][Y][D][T] = 1;

			return array1111;
		}

		[Benchmark]
		public object RandomArray211() {
			var r = new Random(Seed);

			for (int i = 0; i < AccessCount; i++) {
				array211[r.Next(Dim1), r.Next(Dim2)][r.Next(Dim3)][r.Next(Dim4)] = 1;
			}

			return array211;
		}


		[Benchmark]
		public object ConstantArray211() {
			var r = new Random(Seed);

			for (int i = 0; i < AccessCount; i++) {
				array211[1, 1][1][1] = 1;
			}

			return array211;
		}

		[Benchmark]
		public object OneArray211() {
			array211[X, Y][D][T] = 1;

			return array211;
		}

		[Benchmark]
		public int SumRandomValues() {
			var r = new Random(Seed);

			int sum = 0;
			for (int i = 0; i < AccessCount; i++) {
				sum += r.Next(Dim1) + r.Next(Dim2) + r.Next(Dim3) + r.Next(Dim4);
			}

			return sum;
		}

		static void Main(string[] args) {
			BenchmarkRunner.Run<Program>();
		}
	}
}
