﻿using System;
using System.Collections.Generic;

namespace StructsInListNetCore {
	struct Point {
		public int X;
		public int Y;

		public Point(int x, int y) {
			this.X = x;
			this.Y = y;
		}

		public void MultiplyBy(int value) {
			X *= value;
			Y *= value;
		}

		public override string ToString() {
			return $"Point(x: {X}, y: {Y})";
		}

		public void Print() {
			Console.WriteLine(this.ToString());
		}
	}

	class ClassicList<T> {
		T[] data = new T[16];

		public T this[int index] {
			get => data[index];
			set => data[index] = value;
		}
	}

	class Program {
		static void Main(string[] args) {
			Console.WriteLine("+++ Point[] --->");
			Point[] pointArray = new Point[] { new Point(0, 0), new Point(1, 1), new Point(2, 2) };
			pointArray[1].Print();
			pointArray[1].MultiplyBy(1000);
			pointArray[1].Print();

			Console.WriteLine();

			Console.WriteLine("+++ List<Point> --->");
			List<Point> points = new List<Point>();
			points.Add(new Point(0, 0));
			points.Add(new Point(1, 1));
			points.Add(new Point(2, 2));

			points[1].Print();
			points[1].MultiplyBy(1000);
			points[1].Print();

			Console.WriteLine();

			Console.WriteLine("+++ ClassicList<Point> --->");
			ClassicList<Point> cl = new ClassicList<Point>();
			cl[0] = new Point(10, 20);  // calling this setter
			Point p1 = cl[0];           // calling this getter

			p1.MultiplyBy(-1);
			Console.WriteLine($"p1: {p1.ToString()}");
			Console.WriteLine($"cl[0]: {cl[0].ToString()}");

			Console.WriteLine("---");

			cl[0].Print();              // calling this getter
			cl[0].MultiplyBy(1000);     // calling this getter
			cl[0].Print();              // calling this getter

			Console.WriteLine("==============");

			pointArray[0].X = 123;
			pointArray[0].Print();

			// Error: Cannot modify the return value of 'ClassicList<Point>.this[int]' because it is not a variable
			// cl[0].X += 10;

			// Error: Cannot modify the return value of 'List<Point>.this[int]' because it is not a variable
			// points[1].X += 10;

			Console.WriteLine();
			Console.WriteLine("Press ENTER to continue ...");
			Console.ReadLine();

			Console.WriteLine("+++ RefList<Point> --->");
			RefList<Point> rl = new RefList<Point>();
			rl[0] = new Point(10, 20);
			rl[0].Print();
			rl[0].MultiplyBy(1000);
			rl[0].Print();

			Console.WriteLine("Press ENTER to continue ...");
			Console.ReadLine();
			#region Other use cases
			rl[0].X += 10;
			rl[0].Print();

			Console.WriteLine();
			Point p2 = rl[0];
			p2.X *= -1;

			Console.WriteLine($"p2: {p2.ToString()}");
			Console.WriteLine($"rl[0]: {rl[0].ToString()}");

			Console.WriteLine("---");
			ref Point pointRef = ref rl[0];
			pointRef.X *= -2;

			Console.WriteLine($"p2: {p2.ToString()}");
			Console.WriteLine($"rl[0]: {rl[0].ToString()}");
			#endregion
		}
	}

	class RefList<T> {
		T[] data = new T[16];

		public ref T this[int index] {
			get => ref data[index];
		}
	}
}
